anew <<predefine1.f>>
(
Dynamic memory allocation/deallocation is most often used to load files.
Here is a very efficient method to recursively load files, and will
make it unnecessary to use complicated memory management methods.

Assume we have a large memory block, and we use it as TIB.  Forth
interpreter uses three consecutive user variables to manage the
parsing and interpreting of text in TIB/FileBuffer:
	#TIB	Length of TIB or file buffer to be interpreted
	TIB	Beginning of TIB buffer or file buffer
	>IN	Pointer to the next byte to be parsed

FLOAD saves these three pointers by:  
	#TIB 2@ >R >R >IN >R
and reset these pointers to use the space after this TIB/FileBuffer:
	#TIB 2@ + 0 #TIB 2! 0 >IN !
Now, we can read the new file into the new buffer, and store the length
of the new file into #TIB.  EVAL will start interpreting the text in
the new buffer.

When EVAL exits, the text in new buffer is completely interpreted.  We can
restore the interpreter to continue processing remaining text in the old
TIB/File buffer by:
	R> >IN ! R> R> #TIB 2!

This simple method allows files to be FLOAD'ed recursively.  The file buffer
does not have to be very large, because FLOAD's in the same file reuse the 
same buffer area to load and interpret files.

The next file buffer at #TIB 2@ + is always free to be used to load new files
or to store other temporary data.  It is at the same address, as long as this
file is being interpreted.  When this file is completely interpreted, the buffer
it occupies and all the subsequent buffers referred in is file are RELEASED,
and are not valid.  These buffers are overwitten by files whose lengths are
not known.

If we allocate 1 MB for TIB in F#, the system will be greatly simplified.

The other file PreDefine.f contains an alternate implementation, using a file stack.
We can allocate more memory bypushing a new frame on the file stack, and
de-allocate it by popping the frame off the file stack.  The idea is the same,
but the implementation is more complicated.  The #TIB 2@ method is much
simpler and more effective.  It is more suitable for eP32.

)

hidden

CREATE FileBuffer $100000 ALLOT
VARIABLE hFile

( initFileTIB must be executed alone, because it changes TIB address. )
: initFileTIB
	FileBuffer 0 #TIB 2! 
	FileBuffer $100 ERASE ;

: LoadFile ( cstr -- b # )
	1+ ( zstr ) hOpen hFile !
      	hFile @ 
      	#TIB 2@ DROP $10000
	hRead
      	0= ABORT" load file error" 
   	hFile @ CloseHandle 
      	0= ABORT" close file error" 
      	#TIB 2@ DROP ( b )
	#ByteRead ( b # )
   	;

: EVALUATE ( b # -- )
	#TIB 2@ >R >R >IN @ >R
	#TIB 2! 0 >IN !
	EVAL
	R> >IN ! R> R> #TIB 2!
	;		

: FLOAD ( <filename> )
	BL WORD CR ." Loading... " DUP COUNT TYPE
	LoadFile ( b # ) EVALUATE
	;

: \ 	( must redefine so it will not seek end-of-file )
	$0A WORD DROP ; IMMEDIATE

: FILE ( u # -- a n )
	DROP 1+
	LoadFile ( b # ) 
	;

only

